

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
    *                                                                 *
    *   mfcalc 2.0, Extended GNU Bison Example #3                     *
    *   Copyright  2005 Sebastian Pipping <webmaster@hartwork.org>   *
    *                                                                 *
    *   --> http://www.hartwork.org                                   *
    *                                                                 *
    *                                                                 *
    *   This source code is released under GPL.                       *
    *   See GPL.txt for details.                         2005-02-23   *
    *                                                                 *
    \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



%{

/* --------------------------------------------------- */
#include <math.h>		/* Math stuff */
#include <stdio.h>		/* Console stuff */
#include <stdlib.h>		/* <srand> */
#include <time.h>		/* <time> */
#include <ctype.h>		/* <isdigit> */
#include <malloc.h>		/* <malloc> */
#include "calc.h"		/* <symrec> */
#include "functions.h"	/* Calculator extensions */

int yylex (void);
void yyerror (char const *);
/* --------------------------------------------------- */

%}

%union {
	double val;		/* For returning numbers. */
	symrec *tptr;	/* For returning symbol-table pointers. */
	argrec *aptr;
}

%token <val> NUM								/* Simple double precision number. */
%token <tptr> CONST VAR							/* Variables and constants. */
%token <tptr> ZERO ONE TWO THREE MANY			/* Functions. */
%type <val> exp
%type <aptr> args
%right '='
%left '-' '+'
%left '*' '/' '%'
%left NEG		/* negation--unary minus */
%right '^'		/* exponentiation */


%%		/* The grammar follows. */

input:
		/* empty */
	|	input line
;

line:	'\n'
	|	exp '\n'							{ printf ("\t%.10g\n\n", $1); }
	|	error '\n'							{ yyerrok; }
;

exp:	NUM									{ $$ = $1; }
	|	CONST								{ $$ = $1->value.var; }
	|	VAR									{ $$ = $1->value.var; }
	|	VAR '=' exp							{ $$ = $3; $1->value.var = $3; }
	|	ZERO								{ $$ = (*(func_zero)($1->value.fnctptr))(); }
	|	ZERO '(' ')'						{ $$ = (*(func_zero)($1->value.fnctptr))(); }
	|	ONE '(' exp ')'						{ $$ = (*(func_one)($1->value.fnctptr))($3); }
	|	TWO '(' exp ',' exp ')'				{ $$ = (*(func_two)($1->value.fnctptr))($3, $5); }
	|	THREE '(' exp ',' exp ',' exp ')'	{ $$ = (*(func_three)($1->value.fnctptr))($3, $5, $7); }
	|	MANY '(' args ')'					{ $$ = (*(func_many)($1->value.fnctptr))($3); die ($3);}
	|	exp '+' exp							{ $$ = $1 + $3; }
	|	exp '-' exp							{ $$ = $1 - $3; }
	|	exp '*' exp							{ $$ = $1 * $3; }
	|	exp '/' exp							{ $$ = $1 / $3; }
	|	exp '%' exp							{ $$ = OWN_intmod($1,$3); }
	|	'-' exp %prec NEG					{ $$ = -$2; }
	|	exp '^' exp							{ $$ = pow ($1, $3); }
	|	'(' exp ')'							{ $$ = $2; }
;

args:	exp									{ $$ = live ($1); }
	|	exp ',' args						{ $$ = grow ($3, $1)  }
;

%%		/* End of grammar. */


/* --------------------------------------------------- */
/* Called by yyparse on error. */
void
yyerror (char const *s)
{
	printf ("\tERROR: %s\n\n", s);
}
/* --------------------------------------------------- */



struct init_fnct
{
	char const *fname;
	func_t fnct;
};

struct init_fnct const zero[] =
{
	"rnd",		OWN_rnd,
	"rand",		OWN_rnd,
	"time",		OWN_time,

	0,			0
};

struct init_fnct const one[] =
{
	"sin",		sin,
	"cos",		cos,
	"tan",		tan,

	"asin",		asin,
	"arcsin",	asin,
	"acos",		acos,
	"arccos",	acos,
	"atan",		atan,
	"arctan",	atan,

	"sinh",		sinh,
	"cosh",		cosh,
	"tanh",		tanh,

	"ln",		log,
	"log",		log,
	"lg",		log10,
	"log10",	log10,
	"lb",		OWN_ld,
	"ld",		OWN_ld,
	"exp",		exp,

	"sqr",		OWN_sqr,
	"sqrt",		sqrt,
	"abs",		fabs,
	"ceil",		ceil,
	"floor",	floor,
	"int",		floor,

	"deg",		OWN_deg,
	"rad",		OWN_rad,

	"sign",		OWN_sign,
	"sgn",		OWN_sign,

	0,			0
};

struct init_fnct const two[] =
{
	"band",		OWN_band,
	"bor",		OWN_bor,
	"bnot",		OWN_bnot,

	"equal",	OWN_equal,
	"above",	OWN_above,
	"below",	OWN_below,
	"cmp",		OWN_cmp,

	"atan2",	atan2,

	"pow",		pow,

	"mod",		fmod,

	0,			0
};

struct init_fnct const three[] =
{
	"if",		OWN_if,
	"ite",		OWN_if,

	0,			0
};

struct init_fnct const many[] =
{
	"avg",		OWN_avg,
	"mean",		OWN_avg,
	"max",		OWN_max,
	"min",		OWN_min,
	"var",		OWN_var,
	"mse",		OWN_var,
	"stdev",	OWN_stdev,
	"rmse",		OWN_stdev,

	0,			0
};


struct init_const
{
	char const *cname;
	double var;
};

struct init_const const pre_vars[] =
{
	"true",		1.0,
	"false",	0.0,
	"e",		2.718281828,
	"pi",		3.141592653,

	0,			0
};

/* The symbol table: a chain of struct symrec'. */
symrec *sym_table;


/* --------------------------------------------------- */
/* Put arithmetic functions in table. */
void
init_tables (void)
{
	int i;
	symrec *ptr;

	/* Functions */
	for (i = 0; zero[i].fname != 0; i++)
	{
		ptr = putsym (zero[i].fname, ZERO);
		ptr->value.fnctptr = zero[i].fnct;
	}
	for (i = 0; one[i].fname != 0; i++)
	{
		ptr = putsym (one[i].fname, ONE);
		ptr->value.fnctptr = one[i].fnct;
	}
	for (i = 0; two[i].fname != 0; i++)
	{
		ptr = putsym (two[i].fname, TWO);
		ptr->value.fnctptr = two[i].fnct;
	}
	for (i = 0; three[i].fname != 0; i++)
	{
		ptr = putsym (three[i].fname, THREE);
		ptr->value.fnctptr = three[i].fnct;
	}
	for (i = 0; many[i].fname != 0; i++)
	{
		ptr = putsym (many[i].fname, MANY);
		ptr->value.fnctptr = many[i].fnct;
	}

	/* Constants */
	for (i = 0; pre_vars[i].cname != 0; i++)
	{
		ptr = putsym (pre_vars[i].cname, CONST);
		ptr->value.var = pre_vars[i].var;
	}
}
/* --------------------------------------------------- */


/* --------------------------------------------------- */
int
main (void)
{
	srand ((unsigned)time (0));


	printf ("mfcalc 2.0, Extended GNU Bison Example #3\n\n");

	init_tables ();
	return yyparse ();
}
/* --------------------------------------------------- */


/* --------------------------------------------------- */
int
yylex (void)
{
	int c;

	/* Ignore white space, get first nonwhite character. */
	while ((c = getchar ()) == ' ' || c == '\t');
	if (c == EOF)
	{
		return 0;
	}

	/* Char starts a number => parse the number. */
	if (c == '.' || isdigit (c))
	{
		ungetc (c, stdin);
		scanf ("%lf", &yylval.val);
		return NUM;
	}

	/* Char starts an identifier => read the name. */
	if (isalpha (c))
	{
		symrec *s;
		static char *symbuf = 0;
		static int length = 0;
		int i;

		/* Initially make the buffer long enough
		for a 40-character symbol name. */
		if (length == 0)
		{
			length = 40;
			symbuf = (char *)malloc (length + 1);
		}

		i = 0;
		do
		{
			/* If buffer is full, make it bigger. */
			if (i == length)
			{
				length *= 2;
				symbuf = (char *) realloc (symbuf, length + 1);
			}

			/* Add this character to the buffer. */
			symbuf[i++] = c;

			/* Get another character. */
			c = getchar ();
		}

		while (isalnum (c));
		ungetc (c, stdin);
		symbuf[i] = '\0';
		s = getsym (symbuf);
		if (s == 0)
		{
			s = putsym (symbuf, VAR);
		}

		yylval.tptr = s;
		return s->type;
	}

	/* Any other character is a token by itself. */
	return c;
}
/* --------------------------------------------------- */